//
//  Device.cpp
//  InteractiveObjects
//

#include "Device.h"
#include "Evaluator.h"
#include "Context.h"
#include <boost/math/constants/constants.hpp>
#include <boost/range/algorithm/fill.hpp>
#include <boost/make_shared.hpp>
#include <boost/range/algorithm/transform.hpp>
#include <boost/range/algorithm/copy.hpp>
#include <pcl/kdtree/kdtree_flann.h>
#include <pcl/features/normal_3d.h>
#include <pcl/surface/gp3.h>
//#include <pcl/io/vtk_io.h>
#include <pcl/filters/voxel_grid.h>
#include <algorithm>
#include <stdexcept>


namespace
{
        template<typename T>
        inline void create_generator(const xn::NodeInfo& info, const XnMapOutputMode& map_mode, T& generator)
        {
                XnStatus rc = info.GetInstance(generator);
                Evaluator::check_status(rc);
                
                rc = generator.SetMapOutputMode(map_mode);
                Evaluator::check_status(rc);
        }
}

Device::Device(boost::shared_ptr<Context> context)
        : context_(context)
        , image_()
        , depth_()
        , scene_()
        , cloud_()
        , x_factors_()
        , y_factors_()
        , background_cloud_(new pcl::PointCloud<pcl::PointXYZ>)
        , object_cloud_(new pcl::PointCloud<pcl::PointXYZ>)
        , background_mesh_input_generator_(background_cloud_)
        , object_mesh_input_generator_(object_cloud_)
        , empty_generator_()
        , mode_(Default)
        , object_mesh_(boost::make_shared<TriangularMesh>())
        , object_triangles_(boost::make_shared<pcl::PolygonMesh>())
{        
}

void Device::create_factor_tables(int width, int height)
{
        XnFieldOfView fov;
        XnStatus rc = depth_.GetFieldOfView(fov);
        Evaluator::check_status(rc);
        
        const float x_factor = std::tan(fov.fHFOV / 2.0) * 2.0;
        const float y_factor = std::tan(fov.fVFOV / 2.0) * 2.0;

        x_factors_.resize(width);
        for ( int i = 0; i < width; ++i ) {
                x_factors_[i] = x_factor * (0.5f - static_cast<float>(i) / width);
        }
        
        y_factors_.resize(height);
        for ( int i = 0; i < height; ++i ) {
                y_factors_[i] = y_factor * (static_cast<float>(i) / height - 0.5f); 
        }
}



void Device::create_image_generator(const xn::NodeInfo& info, const XnMapOutputMode& map_mode)
{
        create_generator(info, map_mode, image_);
}

void Device::create_depth_generator(const xn::NodeInfo& info, const XnMapOutputMode& map_mode)
{
        create_generator(info, map_mode, depth_);
}

void Device::create_scene_analyzer(const xn::NodeInfo& info, const XnMapOutputMode& map_mode)
{
        create_generator(info, map_mode, scene_);
}

void Device::preprocess(int width, int heigth)
{
        width_ = width;
        height_ = heigth;
        cloud_.resize(width_ * height_ * (sizeof(XnUInt8) * 4 + sizeof(XnFloat) * 3));
        create_factor_tables(width_, height_);
        
        // initialize background_cloud_
        background_cloud_->width = width_;
        background_cloud_->height = height_;
        background_cloud_->resize(background_cloud_->width * background_cloud_->height);
        
        // initialize object_cloud_
        object_cloud_->width = width_;
        object_cloud_->height = height_;
        object_cloud_->resize(object_cloud_->width * object_cloud_->height);
}

namespace
{
        inline void set_colors(const XnRGB24Pixel& c, XnUInt8* dst)
        {
                dst[0] = c.nRed;
                dst[1] = c.nGreen;
                dst[2] = c.nBlue;
        }
       
         inline void set_yellow(XnUInt8* dst)
        {
                dst[0] = 255;
                dst[1] = 255;
                dst[2] = 0;
        }
        
}

void Device::create_cloud()
{
        context_->wait_any_update_all();
                
        xn::ImageMetaData image_meta;
        xn::DepthMetaData depth_meta;
        
        image_.GetMetaData(image_meta);
        depth_.GetMetaData(depth_meta);
        
        if ( mode_ == BackgroundMeshExtraction ) {
                create_cloud(depth_meta, image_meta, boost::ref(background_mesh_input_generator_));
                mode_ = Default;
                boost::packaged_task<boost::shared_ptr<TriangularMesh> > task(boost::bind(&Device::calculate_background_mesh, this));
                future_ = task.get_future();
                boost::thread(boost::move(task)).detach();
        } else {
                create_cloud(depth_meta, image_meta, boost::ref(empty_generator_));
        }
}

namespace
{
	const XnUInt8 Colours[][3] =
	{
		{ 255, 0, 0},
		{ 0, 255, 0},
		{ 0, 0, 255},
		{ 255, 255, 0},
		{ 255, 0, 255},
		{ 0, 255, 255}
	};

        inline void paint_color(const XnUInt16& key, XnUInt8* dst)
        {
                dst[0] = Colours[key][0];
                dst[1] = Colours[key][1];
                dst[2] = Colours[key][2];
        }

        inline void paint_color(XnUInt8* dst)
        {
                dst[0] = 127;
                dst[1] = 127;
                dst[2] = 127;
        }

        const XnUInt32 Step = sizeof(XnUInt8) * 4 + sizeof(XnFloat) * 3;
        
        struct Substractor
        {
                Substractor(XnUInt16 threshold)
                        : threshold_(threshold) {}
                
                XnUInt8 operator()(XnUInt16 d0, XnUInt16 d1) const
                {
                        return std::abs(d0 - d1) < threshold_ ? 0 : 1;
                }
                
                XnUInt16 threshold_;
        };
}

void Device::create_cloud(
        const xn::DepthMetaData&        depth_meta, 
        const xn::ImageMetaData&        image_meta,
        boost::function<void(XnUInt32 k, XnFloat* depth)> fun
)
{
        boost::fill(cloud_, 0);
        const XnDepthPixel* depth_ptr = depth_meta.Data();
        const XnRGB24Pixel* image_ptr = image_meta.RGB24Data();
        
        xn::SceneMetaData scene_meta;
        scene_.GetMetaData(scene_meta);
        const XnLabel* label_ptr = scene_meta.Data();
        XnUInt8* dst_ptr = &cloud_[0];
        const XnUInt32 dst_w = Step * width_;
        background_cloud_size_ = 0;
        object_cloud_size_ = 0;
        for ( XnUInt32 j = 0; j < height_; ++j ) {
                const XnDepthPixel* src_depth = depth_ptr;
                const XnRGB24Pixel* src_image = image_ptr;
                XnUInt8* dst = dst_ptr;
                
                for ( XnUInt32 i = 0; i < width_; ++i, dst += Step ) {
                        const XnDepthPixel& z = src_depth[i];
                        if ( z != 0 ) {

                                // set colors
                                const XnRGB24Pixel& c = src_image[i];
                                set_colors(c, dst);
                                
                                // set coordinates
                                XnFloat* dst_depth = reinterpret_cast<XnFloat*>(dst + 4);
                                convert_to_real_world(i, j, z, dst_depth[0], dst_depth[1], dst_depth[2]);
                        
                                // create an input used to calculate a triangular mesh
                                fun(background_cloud_size_, dst_depth);
                                
                                // set objects
                                const XnUInt16& label = *label_ptr;
                                if ( 0 != label ) {
                                        object_mesh_input_generator_(object_cloud_size_, dst_depth);
                                        ++object_cloud_size_;
                                        paint_color(dst);
                                } 
                                ++background_cloud_size_;
                        }
                        ++label_ptr;
                }
                depth_ptr += width_;
                image_ptr += width_;
                dst_ptr += dst_w;
        }
        if ( object_cloud_size_ ) {
                calculate_object_mesh();
        }
}

void Device::estimate_normals(
        const pcl::PointCloud<pcl::PointXYZ>::Ptr&      cloud,
        pcl::PointCloud<pcl::Normal>::Ptr&              normals
) const
{
        pcl::NormalEstimation<pcl::PointXYZ, pcl::Normal> n;
        pcl::search::KdTree<pcl::PointXYZ>::Ptr tree (new pcl::search::KdTree<pcl::PointXYZ>);
        tree->setInputCloud (cloud);
        n.setInputCloud (cloud);
        n.setSearchMethod (tree);
        n.setKSearch (20);
        n.compute (*normals);
}

namespace
{
        void downsample_cloud(const pcl::PointCloud<pcl::PointXYZ>::Ptr& src, pcl::PointCloud<pcl::PointXYZ>::Ptr& dst)
        {
                pcl::VoxelGrid<pcl::PointXYZ> grid;
                grid.setInputCloud(src);
                grid.setLeafSize(25, 25, 25);
                grid.filter(*dst);
        }
}


void Device::convert_to_polyton_mesh(
        const pcl::PointCloud<pcl::PointXYZ>::Ptr&      ocloud,
        pcl::PolygonMesh&                               triangles
) const
{
        pcl::PointCloud<pcl::PointXYZ>::Ptr cloud(new pcl::PointCloud<pcl::PointXYZ>);
        downsample_cloud(ocloud, cloud);

        pcl::PointCloud<pcl::Normal>::Ptr normals (new pcl::PointCloud<pcl::Normal>);
        estimate_normals(cloud, normals);

        // concatenate the XYZ and normal fields
        // cloud_with_normals = cloud + normals
        pcl::PointCloud<pcl::PointNormal>::Ptr cloud_with_normals (new pcl::PointCloud<pcl::PointNormal>);
        pcl::concatenateFields (*cloud, *normals, *cloud_with_normals);

        // create search tree
        pcl::search::KdTree<pcl::PointNormal>::Ptr tree (new pcl::search::KdTree<pcl::PointNormal>);
        tree->setInputCloud (cloud_with_normals);

        // initialize objects
        pcl::GreedyProjectionTriangulation<pcl::PointNormal> gp;
        
        // set the maximum distance between connected points (maximum edge length)
        gp.setSearchRadius (100);

        // set typical values for the parameters
        gp.setMu (2.5);
        double pi = boost::math::constants::pi<double>();
        gp.setMaximumNearestNeighbors (100);
        gp.setMaximumSurfaceAngle(pi / 4); // 45 degrees
        gp.setMinimumAngle(pi / 18); // 10 degrees
        gp.setMaximumAngle(2 * pi / 3); // 120 degrees
        gp.setNormalConsistency(false);

        // get result
        gp.setInputCloud (cloud_with_normals);
        gp.setSearchMethod (tree);
        gp.reconstruct (triangles);

//        pcl::io::saveVTKFile ("/Users/kumada/Projects/open-ni-tutorial/Triangulation/Triangulation/mesh.vtk", triangles);
}

void Device::extract_ode_data(
        const pcl::PolygonMesh&         triangles,
        std::vector<float>&             vertices,
        std::vector<boost::uint32_t>&   indices
) const
{
        const boost::uint32_t nr_points = triangles.cloud.width * triangles.cloud.height;
        const boost::uint32_t point_size = static_cast<boost::uint32_t>(triangles.cloud.data.size()) / nr_points;
        
        vertices.reserve(3 * nr_points);
        
        for ( boost::uint32_t i = 0; i < nr_points; ++i ) {
                int xyz = 0;
                for ( std::size_t d = 0; d < triangles.cloud.fields.size(); ++d ) {
                        const sensor_msgs::PointField& field = triangles.cloud.fields[d];
                        if ( field.datatype == sensor_msgs::PointField::FLOAT32 ) {
                                const std::string& name = field.name;
                                if ( name == "x" || name == "y" || name == "z" ) {
                                        float value;
                                        memcpy(&value, &triangles.cloud.data[i * point_size + field.offset], sizeof(float));
                                        vertices.push_back(value);
                                        ++xyz;
                                        
                                        if ( xyz == 3 ) {
                                                break;
                                        }
                                }
                        }
                }
        }
        const std::vector<pcl::Vertices>& polygons = triangles.polygons;
        const std::size_t triangle_size = polygons.size();
        indices.reserve(triangle_size * 3);
        for ( std::size_t i = 0; i < triangle_size; ++i ) {
                const std::vector<boost::uint32_t>& vertices = polygons[i].vertices;
                for ( std::size_t j = 0; j < vertices.size(); ++j ) {
                        indices.push_back(vertices[j]);
                }
        }
}

void Device::calculate_object_mesh() const
{
        object_cloud_->resize(object_cloud_size_);
        pcl::PolygonMesh triangles;
        convert_to_polyton_mesh(object_cloud_, triangles);
        
        object_mesh_->vertices_.clear();
        object_mesh_->indices_.clear();
        extract_ode_data(triangles, object_mesh_->vertices_, object_mesh_->indices_);
}

boost::shared_ptr<Device::TriangularMesh> Device::calculate_background_mesh() const
{
        background_cloud_->resize(background_cloud_size_);
        pcl::PolygonMesh triangles;
        convert_to_polyton_mesh(background_cloud_, triangles);
        boost::shared_ptr<TriangularMesh> mesh(boost::make_shared<TriangularMesh>());
        extract_ode_data(triangles, mesh->vertices_, mesh->indices_);
//        std::cout << "mesh->vertices_.size() " << mesh->vertices_.size() << std::endl;
//        std::cout << "mesh->indices_.size() " << mesh->indices_.size() << std::endl;
        return boost::move(mesh);
}

#ifdef UNIT_TEST

#include <boost/test/unit_test.hpp>
#include <iostream>

namespace
{
        int process()
        {
                boost::this_thread::sleep_for(boost::chrono::seconds(3));
                return 100;
        }
        
        void test()
        {
                boost::unique_future<int> future;// = task.get_future();
                        
                for ( int i = 0; i < 2; ++i ) {
                        boost::packaged_task<int> task(boost::bind(&process));
                        future = task.get_future();
                        boost::thread thread(boost::move(task));
                        
                        while ( !future.is_ready() ) {
                                //std::cout << "sss\n";
                        }
                        
                        boost::unique_future<int> f(boost::move(future));
                        std::cout << f.get() << std::endl;
                        
                        std::cout << f.valid() << std::endl;
                        std::cout << "ttt\n";
                        std::cout << future.valid() << std::endl;
                }
        }
}

BOOST_AUTO_TEST_CASE(TEST_Device)
{
        test();
        std::cout << "TEST_Device\n";
}

#endif // UNIT_TEST