//
//  Device.h
//  InteractiveObjects
//

#ifndef __OpenNiMultiViewer__Device__
#define __OpenNiMultiViewer__Device__

#include <boost/thread/future.hpp>
#include <boost/function.hpp>
#include <boost/noncopyable.hpp>
#include <boost/thread.hpp>
#include <boost/array.hpp>
#include <pcl/point_types.h>
#include <pcl/surface/gp3.h>
#include <XnCppWrapper.h>
#include <vector>
#include <string>

class Context;
namespace pcl
{
        class PolygonMesh;
}

class Device : private boost::noncopyable
{
public:
        enum Mode
        {
                Default,
                BackgroundMeshExtraction,
        };
        
        struct TriangularMesh
        {
                std::vector<float>              vertices_;
                std::vector<boost::uint32_t>    indices_;
                
                bool empty() const
                {
                        return vertices_.empty() || indices_.empty();
                }
                
        };
        
        explicit Device(boost::shared_ptr<Context> context);
        void create_image_generator(const xn::NodeInfo& info, const XnMapOutputMode& map_mode);
        void create_depth_generator(const xn::NodeInfo& info, const XnMapOutputMode& map_mode);
        void create_scene_analyzer(const xn::NodeInfo& info, const XnMapOutputMode& map_mode);
        
        boost::shared_ptr<TriangularMesh> get_background_mesh()
        {
                boost::unique_future<boost::shared_ptr<TriangularMesh> > f(boost::move(future_));
                return boost::move(f.get());
        }
        
        const boost::shared_ptr<TriangularMesh>& get_object_mesh() const
        {
                return object_mesh_;
        }
        
        bool is_ready_for_background_mesh() const
        {
                return future_.valid() && future_.is_ready();
        }
        
        const xn::ImageGenerator& get_image() const
        {
                return image_;
        }
        
        const xn::DepthGenerator& get_depth() const
        {
                return depth_;
        }
        
        void preprocess(int width, int height);
        int get_cloud_size() const
        {
                return cloud_.size() / (sizeof(XnUInt8) * 4 + sizeof(XnFloat) * 3);
        }
        
        const std::vector<XnUInt8>& get_cloud() const
        {
                return cloud_;
        }
        void create_cloud();
        
        void set_image_info_string(const std::string& info)
        {
                image_info_string_ = info;
        }
        
        void set_depth_info_string(const std::string& info)
        {
                depth_info_string_ = info;
        }

        void set_scene_info_string(const std::string& info)
        {
                scene_info_string_ = info;
        }
        
        const std::string& get_image_info_string() const
        {
                return image_info_string_;
        }
        
        const std::string& get_depth_info_string() const
        {
                return depth_info_string_;
        }
        
        const std::string& get_scene_info_string() const
        {
                return scene_info_string_;
        }

        void set_mode(Mode mode)
        {
                mode_ = mode;
        }
        
        const Mode get_mode() const
        {
                return mode_;
        }
        
private:
        struct MeshInputGenerator
        {
                MeshInputGenerator(pcl::PointCloud<pcl::PointXYZ>::Ptr cloud)
                        : cloud_(cloud) {}
                        
                void operator()(XnUInt32 k, XnFloat* depth) const
                {
                        pcl::PointXYZ& p = cloud_->points[k];
                        p.x = depth[0];
                        p.y = depth[1];
                        p.z = depth[2];
                }
                
                pcl::PointCloud<pcl::PointXYZ>::Ptr cloud_;
        };
        
        struct EmptyGenerator
        {
                void operator()(XnUInt32 k, XnFloat* depth) const
                {
                        // do nothing
                }
        };

        boost::shared_ptr<Context> context_;
        xn::ImageGenerator      image_;
        xn::DepthGenerator      depth_;
        xn::SceneAnalyzer       scene_;
        std::vector<XnUInt8>    cloud_;
        
        int width_;
        int height_;
        
        std::vector<float> x_factors_;
        std::vector<float> y_factors_;
        
        std::string image_info_string_;
        std::string depth_info_string_;
        std::string scene_info_string_;
        
        // background
        pcl::PointCloud<pcl::PointXYZ>::Ptr background_cloud_;
        int background_cloud_size_;
        
        // object
        pcl::PointCloud<pcl::PointXYZ>::Ptr     object_cloud_;
        boost::shared_ptr<TriangularMesh>       object_mesh_;
        pcl::PolygonMesh::Ptr                   object_triangles_;
        int                                     object_cloud_size_;
        
        Mode mode_;
        
        MeshInputGenerator      background_mesh_input_generator_;
        MeshInputGenerator      object_mesh_input_generator_;
        EmptyGenerator          empty_generator_;
        
        boost::unique_future<boost::shared_ptr<TriangularMesh> > future_;
        
        void create_factor_tables(int width, int height);
        
        void convert_to_real_world(
                const XnUInt32&         i, 
                const XnUInt32&         j, 
                const XnDepthPixel&     depth, 
                float&                  x, 
                float&                  y, 
                float&                  z
        ) const
        {
                z = -static_cast<float>(depth);
                x = z * x_factors_[i];
                y = z * y_factors_[j];
        }
        
        void create_cloud(
                const xn::DepthMetaData&        depth_meta, 
                const xn::ImageMetaData&        image_meta,
                boost::function<void(XnUInt32 k, XnFloat* depth)> fun
        );
        
        void convert_to_polyton_mesh(
                const pcl::PointCloud<pcl::PointXYZ>::Ptr&      cloud,
                pcl::PolygonMesh&                               triangles
        ) const;
        
        void estimate_normals(
                const pcl::PointCloud<pcl::PointXYZ>::Ptr&      cloud,
                pcl::PointCloud<pcl::Normal>::Ptr&              normals
        ) const;

        void calculate_object_mesh() const;
        boost::shared_ptr<TriangularMesh> calculate_background_mesh() const;

        void extract_ode_data(
                const pcl::PolygonMesh&         triangles,
                std::vector<float>&             vertices,
                std::vector<boost::uint32_t>&   indices
        ) const;
};

#endif /* defined(__OpenNiMultiViewer__Device__) */
