//
//  DeviceEnumerator.cpp
//  InteractiveObjects
//

#include "DeviceEnumerator.h"
#include <boost/range/algorithm/for_each.hpp>
#include <stdexcept>
#include <sstream>
#include "Device.h"
#include "Evaluator.h"
#include "Context.h"

DeviceEnumerator::DeviceEnumerator()
        : devices_()
{
}

const std::string DeviceEnumerator::get_info_string(const xn::NodeInfo& node) const
{
        std::ostringstream oss;
        oss << xnProductionNodeTypeToString(node.GetDescription().Type)
                << ", "
                << node.GetCreationInfo() << ", "
                << node.GetInstanceName() << ", "
                << node.GetDescription().strName << ", "
                << node.GetDescription().strVendor;
        return oss.str();
}

void DeviceEnumerator::enumerate_production_trees(boost::shared_ptr<Context> context, XnProductionNodeType type)
{
        std::cout << "xn::Context::EnumerateProductionTrees ... ";
        xn::NodeInfoList nodes;
        context->enumerate_production_trees(nodes, type);
        for ( xn::NodeInfoList::Iterator it = nodes.Begin (); it != nodes.End (); ++it ) {
                xn::NodeInfo info = *it;
                std::cout << " " << get_info_string(info) << std::endl;
                context->create_production_tree(info);
        }
}

void DeviceEnumerator::enumerate_production_trees(boost::shared_ptr<Context> context)
{
        enumerate_production_trees(context, XN_NODE_TYPE_DEVICE);
        enumerate_production_trees(context, XN_NODE_TYPE_IMAGE);
        enumerate_production_trees(context, XN_NODE_TYPE_DEPTH);
        enumerate_production_trees(context, XN_NODE_TYPE_SCENE);
}

namespace
{
        struct DeviceInfo
        {
                void operator()(const std::map<int, boost::shared_ptr<Device> >::value_type& pair) const
                {
                        std::cout << "[" << pair.first + 1 << "] " << pair.second->get_image_info_string() << std::endl;
                        std::cout << "[" << pair.first + 1 << "] " << pair.second->get_depth_info_string() << std::endl;
                        std::cout << "[" << pair.first + 1 << "] " << pair.second->get_scene_info_string() << std::endl;
                }
        };
}

void DeviceEnumerator::enumerate_existing_nodes(boost::shared_ptr<Context> context, const XnMapOutputMode& map_mode)
{
        xn::NodeInfoList node_info_list;
        std::cout << "xn::Context::EnumerateExistingNodes ... ";
        context->enumerate_existing_nodes(node_info_list);
        
        std::cout << "Success" << std::endl;
        
        devices_.clear();
        xn::NodeInfoList::Iterator first = node_info_list.Begin();
        xn::NodeInfoList::Iterator last = node_info_list.End();
        while ( first != last ) {
                const xn::NodeInfo& info = *first;
                const std::string& name = info.GetInstanceName();
                int no = *name.rbegin() - '1';
                if ( !devices_[no] ) {
                        devices_[no] = boost::shared_ptr<Device>(new Device(context));
                }
                
                if ( info.GetDescription().Type == XN_NODE_TYPE_IMAGE ) {
                        devices_[no]->create_image_generator(info, map_mode);
                        devices_[no]->set_image_info_string(get_info_string(info));
                } else if ( info.GetDescription().Type == XN_NODE_TYPE_DEPTH ) {
                        devices_[no]->create_depth_generator(info, map_mode);
                        devices_[no]->set_depth_info_string(get_info_string(info));
                } else if ( info.GetDescription().Type == XN_NODE_TYPE_SCENE ) {
                        devices_[no]->create_scene_analyzer(info, map_mode);
                        devices_[no]->set_scene_info_string(get_info_string(info));
                } else {
                        //
                }
                ++first;
        }
        
        boost::for_each(devices_, DeviceInfo());
}