//
//  PhysicalSimulator.cpp
//  InteractiveObjects
//

#include "PhysicalSimulator.h"
#include <boost/move/move.hpp>
#include <boost/make_shared.hpp>
#include <boost/math/constants/constants.hpp>
#include <pcl/kdtree/kdtree_flann.h>
#include <pcl/features/normal_3d.h>
#include <boost/random.hpp>
#include <ctime>

PhysicalSimulator::PhysicalSimulator()
        : contacts_(MaxNumOfContacts)
        , background_mesh_(0)
        , object_mesh_(0)
{
        dInitODE();
}

PhysicalSimulator::~PhysicalSimulator()
{
        dCloseODE();
}

void PhysicalSimulator::collide(void(*callback)(void* data, dGeomID o1, dGeomID o2))
{
        space_.collide(0, callback);
}

void PhysicalSimulator::near_callback(void* data, dGeomID o1, dGeomID o2)
{
        int n = dCollide(o1, o2, MaxNumOfContacts, &contacts_[0].geom, sizeof(dContact));
        for ( int i = 0; i < n; ++i ) {
                
                dContact& contact = contacts_[i];
                contact.surface.mode = dContactBounce | dContactSoftERP | dContactSoftCFM;
                contact.surface.mu = dInfinity;
                contact.surface.mu2 = 0;//dInfinity;
                contact.surface.bounce = 0.5;
//                contact.surface.bounce_vel = 0.1;
                contact.surface.soft_cfm = 0.0;
                contact.surface.soft_erp = 1.0;
                dContactJoint joint;
                joint.create(world_, joint_group_, &contact);
                joint.attach(dGeomGetBody(contact.geom.g1), dGeomGetBody(contact.geom.g2));
        }
}

void PhysicalSimulator::clear_balls()
{
        for ( std::size_t i = 0; i < balls_.size(); ++i ) {
                boost::shared_ptr<Ball>& ball = balls_[i];
                ball->geom_.destroy();
        }
        balls_.clear();
}

void PhysicalSimulator::clear_cylinders()
{
        for ( std::size_t i = 0; i < cylinders_.size(); ++i ) {
                boost::shared_ptr<dCylinder>& cylinder = cylinders_[i];
                cylinder->destroy();
        }
        cylinders_.clear();
}

void PhysicalSimulator::create_world()
{
        const dReal gravity[] = {0, -1, 0};
        world_.setGravity(gravity);
}

void PhysicalSimulator::create_cylinders()
{
        dReal radius = 80;
        dReal length = 150;
        dReal center = -460 + length / 2;
        dMatrix3 initial_rotation;
        const double Pi = boost::math::constants::pi<double>();
        dRFromAxisAndAngle(initial_rotation, 1.0, 0.0, 0.0, Pi / 2.0);
  
        clear_cylinders();
        dReal step = 300;
        create_cylinder(initial_rotation, 100 + 1.5 * step, center + 50, -1200, radius, length);
        create_cylinder(initial_rotation, 200, center, -1100, radius, length);
        create_cylinder(initial_rotation, 100 + step + 10, center + 50, -1100, radius, length);
        create_cylinder(initial_rotation, 100 - step / 2, center + 50, -1200, radius, length);
        create_cylinder(initial_rotation, 100 - step - 100, center + 50, -1100, radius, length);
        create_cylinder(initial_rotation, 100 - step - 300, center + 50, -800 - step, radius, length);
}

void PhysicalSimulator::create_cylinder(
        const dMatrix3& initial_rotation,
        dReal x,
        dReal y,
        dReal z,
        dReal radius,
        dReal length
)
{
        boost::shared_ptr<dCylinder> cylinder = boost::make_shared<dCylinder>();
        cylinder->create(space_, radius, length);
        cylinder->setPosition(x, y, z);
        cylinder->setRotation(&initial_rotation[0]);
        cylinders_.push_back(boost::move(cylinder));
}

void PhysicalSimulator::create_balls()
{
        dMass m;
        m.setZero();
        const dReal mass = 100;
        dReal radius = 80;
        m.setSphereTotal(mass, radius);
        const dReal initial_position[] = {200, -85, -1900};

        const int NumOfBalls = 15;
        balls_.clear();
        boost::mt19937 gen(static_cast<int>(time(0)));
        int max_number = 100;
        boost::uniform_int<> dst(1, max_number);
        boost::variate_generator<boost::mt19937, boost::uniform_int<> > rand(gen, dst);
        for ( std::size_t i = 0; i < NumOfBalls; ++i ) {
                boost::shared_ptr<Ball> ball = boost::make_shared<Ball>();
                
                ball->body_.create(world_);
                ball->body_.setMass(&m);
                
                if ( i < 5 ) {
                        ball->body_.setPosition(
                                initial_position[0] + radius * 2 * i - radius * (NumOfBalls / 2- 1),
                                initial_position[1],
                                initial_position[2]
                        );
                } else if ( i < 10 ) {
                        ball->body_.setPosition(
                                initial_position[0] + radius * 2 * (i - 5) - radius * (NumOfBalls / 2 - 1),
                                initial_position[1],
                                initial_position[2] + radius * 2
                        );
                } else {
                        ball->body_.setPosition(
                                initial_position[0] + radius * 2 * (i - 10) - radius * (NumOfBalls / 2 - 1),
                                initial_position[1],
                                initial_position[2] + radius * 4
                        );
                }

                // The geometry(ball.geom) is associated with the rigid body(ball.body).
                ball->geom_.create(space_, radius);
                ball->geom_.setBody(ball->body_);
                balls_.push_back(boost::move(ball));
        }
}

void PhysicalSimulator::register_background_mesh(const boost::shared_ptr<Device::TriangularMesh>& mesh)
{
        dTriMeshDataID data = dGeomTriMeshDataCreate();
        dGeomTriMeshDataBuildSingle(
                data,
                &(mesh->vertices_[0]), 3 * sizeof(float), mesh->vertices_.size() / 3, &(mesh->indices_[0]),
                static_cast<int>(mesh->indices_.size()),
                3 * sizeof(dTriIndex)
        );
        if ( background_mesh_)  {
                dGeomDestroy(background_mesh_);
        }
        background_mesh_ = dCreateTriMesh(space_, data, 0, 0, 0);
}

void PhysicalSimulator::register_object_mesh(const boost::shared_ptr<Device::TriangularMesh>& mesh)
{
        dTriMeshDataID data = dGeomTriMeshDataCreate();
        dGeomTriMeshDataBuildSingle(
                data,
                &(mesh->vertices_[0]), 3 * sizeof(float), mesh->vertices_.size() / 3, &(mesh->indices_[0]),
                static_cast<int>(mesh->indices_.size()),
                3 * sizeof(dTriIndex)
        );
        if ( object_mesh_)  {
                dGeomDestroy(object_mesh_);
        }
        object_mesh_ = dCreateTriMesh(space_, data, 0, 0, 0);
        
}